# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, Iterable, Iterator, Optional, TypeAlias, TypeVar

from .config import ParallelConfig, ensure_parallel_init, get_parallel_config

M = TypeVar("M")
A = TypeVar("A")
R = TypeVar("R")
InvokeOp: TypeAlias = Callable[[M, A], R]


def invoker(
    model: M,
    func: InvokeOp[M, A, R],
    n_jobs: Optional[int] = None,
    *,
    worker_parallel: ParallelConfig | None = None,
) -> ModelOpInvoker[A, R]:
    """
    Get an appropriate invoker for performing operations on ``model``.

    Args:
        model:
            The model object on which to perform operations.
        func:
            The function to call.  The function must be pickleable.
        n_jobs:
            The number of processes to use for parallel operations.  If ``None``, will
            call :func:`proc_count` with a maximum default process count of 4.
        worker_parallel:
            A parallel configuration for subprocess workers.

    Returns:
        ModelOpInvoker:
            An invoker to perform operations on the model.
    """
    ensure_parallel_init()
    if n_jobs is None:
        n_jobs = get_parallel_config().processes

    if n_jobs == 1:
        from .sequential import InProcessOpInvoker

        return InProcessOpInvoker(model, func)
    else:
        from .pool import ProcessPoolOpInvoker

        return ProcessPoolOpInvoker(model, func, n_jobs, worker_parallel=worker_parallel)


class ModelOpInvoker(ABC, Generic[A, R]):
    """
    Interface for invoking operations on a model, possibly in parallel.  The operation
    invoker is configured with a model and a function to apply, and applies that function
    to the arguments supplied in `map`.  Child process invokers also route logging messages
    to the parent process, so logging works even with multiprocessing.

    An invoker is a context manager that calls :meth:`shutdown` when exited.
    """

    @abstractmethod
    def map(self, tasks: Iterable[A]) -> Iterator[R]:
        """
        Apply the configured function to the model and iterables.  This is like
        :func:`map`, except it supplies the invoker's model as the first object
        to ``func``.

        Args:
            iterables: Iterables of arguments to provide to the function.

        Returns:
            iterable: An iterable of the results.
        """
        pass

    def shutdown(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args: Any):
        self.shutdown()



"""
Top-level LensKit functions for recommender operations.
"""

from __future__ import annotations

from lenskit.data import ItemList, QueryInput
from lenskit.data.types import IDSequence
from lenskit.pipeline import Pipeline


def recommend(
    pipeline: Pipeline,
    query: QueryInput,
    n: int | None = None,
    items: ItemList | IDSequence | None = None,
    *,
    component: str = "recommender",
) -> ItemList:
    """
    Generate recommendations for a user or query.  This calls the specified
    pipeline component (the ``'recommender'`` by default) and returns the
    resulting item list.

    Stability:
        full
    Args:
        pipeline:
            The pipeline to run.
        query:
            The user ID or other query data for the recommendation.
        n:
            The number of items to recommend.
        items:
            The candidate items, or ``None`` to use the pipeline's default
            candidate selector.
        component:
            The name of the component implementing the recommender.

    Returns:
        The recommended items as an ordered item list.
    """

    node = pipeline.node(component)
    if items is not None and not isinstance(items, ItemList):
        items = ItemList(items)
    res = pipeline.run(node, query=query, n=n, items=items)
    if not isinstance(res, ItemList):
        raise TypeError("recommender pipeline did not return an item list")

    return res


def score(
    pipeline: Pipeline,
    query: QueryInput,
    items: ItemList | IDSequence,
    *,
    component: str = "scorer",
) -> ItemList:
    """
    Score items with respect to a user or query.  This calls the specified
    pipeline component (the ``'scorer'`` by default) and returns the resulting
    item list.

    Stability:
        Full
    Args:
        pipeline:
            The pipeline to run.
        query:
            The user ID or other query data for the recommendation.
        items:
            The candidate items, or ``None`` to use the pipeline's default
            candidate selector.
        component:
            The name of the component implementing the scorer.
    Returns:
        The items with scores in the ``score`` field.
    """

    node = pipeline.node(component)
    if items is not None and not isinstance(items, ItemList):
        items = ItemList(items)
    res = pipeline.run(node, query=query, items=items)
    if not isinstance(res, ItemList):
        raise TypeError("scorer pipeline did not return an item list")

    return res


def predict(
    pipeline: Pipeline,
    query: QueryInput,
    items: ItemList | IDSequence,
    *,
    component: str = "rating-predictor",
) -> ItemList:
    """
    Predict ratings for items.  This is exactly like :func:`score`, except it
    defaults to the ``'rating-predictor'`` component.  In a standard pipeline,
    the rating predictor may have additional configuration such as fallbacks or
    transformations to ensure every item is scored and the scores are valid
    rating predictions; the scorer typically returns raw scores.

    Stability:
        Full
    """

    return score(pipeline, query, items, component=component)



from lenskit.data import ItemList

from .._base import GlobalMetric, ListMetric, Metric

__all__ = ["Metric", "ListMetric", "GlobalMetric", "RankingMetricBase"]


class RankingMetricBase(Metric):
    """
    Base class for most ranking metrics, implementing a ``k`` parameter for
    truncation.

    Args:
        k:
            Specify the length cutoff for rankings. Rankings longer than this
            will be truncated prior to measurement.

    Stability:
        Caller
    """

    k: int | None = None
    "The maximum length of rankings to consider."

    def __init__(self, k: int | None = None):
        if k is not None and k < 0:
            raise ValueError("k must be positive or None")
        self.k = k

    def truncate(self, items: ItemList):
        """
        Truncate an item list if it is longer than :attr:`k`.
        """
        if self.k is not None:
            if not items.ordered:
                raise ValueError("top-k filtering requires ordered list")
            if len(items) > self.k:
                return items[: self.k]

        return items



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
"""
Bias scoring model.
"""

from __future__ import annotations

import logging
from collections.abc import Container
from dataclasses import dataclass
from typing import Annotated, Literal

import numpy as np
import torch
from pydantic import BaseModel, NonNegativeFloat, PlainSerializer
from typing_extensions import Self, TypeAlias, overload

from lenskit.data import ID, Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.pipeline.components import Component
from lenskit.training import Trainable, TrainingOptions

_logger = logging.getLogger(__name__)
BiasEntity: TypeAlias = Literal["user", "item"]
Damping: TypeAlias = float | dict[BiasEntity, float]


@dataclass
class BiasModel:
    """
    User-item bias models learned from rating data.  The :class:`BiasScorer`
    class uses this model to score items in a pipeline; the model is reusable
    in other components that need user-item bias models.

    This implements the following model:

    .. math::
       b_{ui} = b_g + b_i + b_u

    where :math:`b_g` is the global bias (global mean rating), :math:`b_i` is
    item bias, and :math:`b_u` is the user bias.  With the provided damping
    values :math:`\\beta_{\\mathrm{u}}` and :math:`\\beta_{\\mathrm{i}}`, they
    are computed as follows:

    .. math::
       \\begin{align*}
       b_g & = \\frac{\\sum_{r_{ui} \\in R} r_{ui}}{|R|} &
       b_i & = \\frac{\\sum_{r_{ui} \\in R_i} (r_{ui} - b_g)}{|R_i| + \\beta_{\\mathrm{i}}} &
       b_u & = \\frac{\\sum_{r_{ui} \\in R_u} (r_{ui} - b_g - b_i)}{|R_u| + \\beta_{\\mathrm{u}}}
       \\end{align*}

    The damping values can be interpreted as the number of default (mean)
    ratings to assume *a priori* for each user or item, damping low-information
    users and items towards a mean instead of permitting them to take on extreme
    values based on few ratings.

    Stability:
        Caller
    """

    damping: Damping
    "The mean damping terms."

    global_bias: float
    "The global bias term."

    items: Vocabulary | None = None
    "Vocabulary of items."
    item_biases: np.ndarray[int, np.dtype[np.float32]] | None = None
    "The item offsets (:math:`b_i` values)."

    users: Vocabulary | None = None
    "Vocabulary of users."
    user_biases: np.ndarray[int, np.dtype[np.float32]] | None = None
    "The user offsets (:math:`b_u` values)."

    @classmethod
    def learn(
        cls,
        data: Dataset,
        damping: Damping | tuple[float, float] = 0.0,
        *,
        entities: Container[BiasEntity] = frozenset({"user", "item"}),
    ) -> Self:
        """
        Learn a bias model and its parameters from a dataset.

        Args:
            data:
                The dataset from which to learn the bias model.
            damping:
                Bayesian damping to apply to computed biases.  Either a number, to
                damp both user and item biases the same amount, or a (user,item)
                tuple providing separate damping values.
            items:
                Whether to compute item biases
            users:
                Whether to compute user biases
        """

        if isinstance(damping, tuple):
            damping = {"user": damping[0], "item": damping[1]}
        _logger.info("building bias model for %d ratings", data.interaction_count)
        ratings = data.interaction_matrix("scipy", layout="coo", field="rating")
        nrows, ncols = ratings.shape  # type: ignore

        g_bias = float(np.mean(ratings.data))
        _logger.info("global mean: %.3f", g_bias)

        model = cls(damping, g_bias)

        centered = ratings.data - g_bias
        if np.allclose(centered, 0):
            _logger.warning("mean-centered ratings are all 0, bias probably meaningless")

        if "item" in entities:
            counts = np.full(ncols, entity_damping(damping, "item"))
            sums = np.zeros(ncols)
            np.add.at(counts, ratings.col, 1)
            np.add.at(sums, ratings.col, centered)

            # store 0 offsets
            i_bias = np.zeros(ncols, dtype=np.float32)
            np.divide(sums, counts, out=i_bias, where=counts > 0)

            model.items = data.items.copy()
            model.item_biases = i_bias
            centered -= i_bias[ratings.col]
            _logger.info("computed biases for %d items", len(i_bias))

        if "user" in entities:
            counts = np.full(nrows, entity_damping(damping, "user"))
            sums = np.zeros(nrows)
            np.add.at(counts, ratings.row, 1)
            np.add.at(sums, ratings.row, centered)

            u_bias = np.zeros(nrows, dtype=np.float32)
            np.divide(sums, counts, out=u_bias, where=counts > 0)

            model.users = data.users.copy()
            model.user_biases = u_bias
            _logger.info("computed biases for %d users", len(u_bias))

        return model

    @overload
    def compute_for_items(
        self,
        items: ItemList,
        user_id: ID | None = None,
        user_items: ItemList | None = None,
    ) -> tuple[np.ndarray[tuple[int], np.dtype[np.float32]], float]: ...
    @overload
    def compute_for_items(
        self,
        items: ItemList,
        *,
        bias: float,
    ) -> np.ndarray[tuple[int], np.dtype[np.float32]]: ...
    def compute_for_items(
        self,
        items: ItemList,
        user_id: ID | None = None,
        user_items: ItemList | None = None,
        *,
        bias: float | None = None,
    ):
        """
        Compute the personalized biases for a set of itemsm and optionally a
        user.  The user can be specified either by their identifier or by a list
        of ratings.

        Args:
            items:
                The items to score.
            user:
                The user identifier.
            user_items:
                The user's items, with ratings (takes precedence over ``user``
                if both are supplied).  If the supplied list does not have a
                ``rating`` field, it is ignored.
            bias:
                A pre-computed user bias.
        Returns:
            A tuple of the overall bias scores for the specified items and user,
            and the user's bias (needed to de-normalize scores efficiently
            later).  If a user bias is provided instead of user information,
            only the composite bias scores are returned.
        """
        n = len(items)
        scores = np.full(n, self.global_bias, dtype=np.float32)

        if self.item_biases is not None:
            assert self.items is not None
            idxes = items.numbers(vocabulary=self.items, missing="negative")
            mask = idxes >= 0
            scores[mask] += self.item_biases[idxes[mask]]

        if bias is not None:
            return scores + bias

        ratings = None
        if user_items is not None:
            ratings = user_items.field("rating")

        user_bias = 0.0
        if self.users is not None:
            assert self.user_biases is not None

            if ratings is not None:
                assert user_items is not None

                uoff = ratings - self.global_bias

                if self.item_biases is not None:
                    r_idxes = user_items.numbers(vocabulary=self.items, missing="negative")
                    r_mask = r_idxes >= 0
                    uoff[r_mask] -= self.item_biases[r_idxes[r_mask]]

                user_bias = np.sum(uoff) / (
                    np.sum(np.isfinite(uoff)) + entity_damping(self.damping, "user")
                )
                scores += user_bias

            elif user_id is not None:
                uno = self.users.number(user_id, missing="none")
                if uno is not None:
                    user_bias = self.user_biases[uno]
                    _logger.debug("using mean(user %s) = %.3f", user_id, user_bias)
                    scores += user_bias

        return scores, user_bias

    def transform_matrix(self, matrix: torch.Tensor):
        """
        Transform a sparse ratings matrix by subtracting biases.
        """
        if not matrix.is_sparse:
            raise TypeError("matrix is not sparse COO")

        indices = matrix.indices()
        unos = indices[0, :]
        inos = indices[1, :]
        values = matrix.values() - self.global_bias
        if self.item_biases is not None:
            values.subtract_(torch.from_numpy(self.item_biases)[inos])
        if self.user_biases is not None:
            values.subtract_(torch.from_numpy(self.user_biases)[unos])
        return torch.sparse_coo_tensor(indices, values, size=matrix.size())


class BiasConfig(BaseModel, extra="forbid"):
    """
    Configuration for :class:`BiasScorer`.
    """

    entities: Annotated[
        set[Literal["user", "item"]], PlainSerializer(lambda s: sorted(s), return_type=list[str])
    ] = {"user", "item"}
    """
    The entities to compute biases for, in addition to global bais.  Defaults to
    users and items.
    """
    damping: NonNegativeFloat | dict[Literal["user", "item"], NonNegativeFloat] = 0.0

    def entity_damping(self, entity: Literal["user", "item"]) -> float:
        """
        Look up the damping for a particular entity type.
        """
        return entity_damping(self.damping, entity)


class BiasScorer(Component[ItemList], Trainable):
    """
    A user-item bias rating prediction model.  This component uses
    :class:`BiasModel` to predict ratings for users and items.

    Args:
        config:
            The component configuration.

    Stability:
        Caller
    """

    config: BiasConfig
    model_: BiasModel

    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        """
        Train the bias model on some rating data.

        Args:
            ratings:
                The training data (must have ratings).

        Returns:
            The trained bias object.
        """
        if hasattr(self, "model_") and not options.retrain:
            return

        self.model_ = BiasModel.learn(data, self.config.damping, entities=self.config.entities)

    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        """
        Compute predictions for a user and items.  Unknown users and items are
        assumed to have zero bias.

        Args:
            query:
                The recommendation query.  If the query has an item list with
                ratings, those ratings are used to compute a new bias instead of
                using the user's recorded bias.
            items:
                The items to score.
        Returns:
            Scores for `items`.
        """
        query = RecQuery.create(query)

        scores, _bias = self.model_.compute_for_items(items, query.user_id, query.user_items)
        return ItemList(items, scores=scores)


def entity_damping(damping: Damping, entity: BiasEntity) -> float:
    """
    Look up the damping for a particular entity type.
    """
    if isinstance(damping, dict):
        return damping.get(entity, 0.0)
    else:
        return damping



from __future__ import annotations

import logging

import numpy as np

from lenskit.data.items import ItemList
from lenskit.pipeline import Component

_logger = logging.getLogger(__name__)


class FallbackScorer(Component[ItemList]):
    """
    Scoring component that fills in missing scores using a fallback.

    Stability:
        Caller
    """

    config: None

    def __call__(self, scores: ItemList, backup: ItemList) -> ItemList:
        s = scores.scores()
        if s is None:
            return backup

        s = s.copy()
        missing = np.isnan(s)
        if not np.any(missing):
            return scores

        bs = backup.scores("pandas", index="ids")
        if bs is None:
            return scores

        s[missing] = bs.reindex(scores.ids()[missing]).values
        return ItemList(scores, scores=s)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import numpy as np
import torch


@dataclass
class TorchUserItemTable:
    """
    Table of user-item interaction data represented as PyTorch tensors.

    Stability:
        Caller
    """

    user_nums: torch.Tensor
    """
    User numbers (0-based contiguous integers, see :ref:`data-identifiers`).
    """
    item_nums: torch.Tensor
    """
    Item numbers (0-based contiguous integers, see :ref:`data-identifiers`).
    """
    ratings: Optional[torch.Tensor] = None
    """
    Ratings for the items.
    """
    timestamps: Optional[torch.Tensor] = None
    """
    Timestamps for recorded user-item interactions.
    """


@dataclass
class NumpyUserItemTable:
    """
    Table of user-item interaction data represented as NumPy arrays.

    Stability:
        Caller
    """

    user_nums: np.ndarray[int, np.dtype[np.int32]]
    """
    User numbers (0-based contiguous integers, see :ref:`data-identifiers`).
    """
    item_nums: np.ndarray[int, np.dtype[np.int32]]
    """
    Item numbers (0-based contiguous integers, see :ref:`data-identifiers`).
    """
    ratings: Optional[np.ndarray[int, np.dtype[np.float32]]] = None
    """
    Ratings for the items.
    """
    timestamps: Optional[np.ndarray[int, np.dtype[np.int64]]] = None
    """
    Timestamps for recorded user-item interactions.
    """



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import argparse
import logging
import sys
from pathlib import Path
from urllib.request import urlopen

_log = logging.getLogger("lenskit.data.fetch")

ML_LOC = "http://files.grouplens.org/datasets/movielens/"


def fetch_ml(name: str, base_dir: Path):
    """
    Fetch a MovieLens dataset.  The followings names are recognized:

    . ml-100k
    . ml-1m
    . ml-10m
    . ml-20m
    . ml-25m
    . ml-latest
    . ml-latest-small

    Args:
        name:
            The name of the dataset.
        base_dir:
            The base directory into which data should be downloaded.
    """
    zipname = f"{name}.zip"
    zipfile = base_dir / zipname
    zipurl = ML_LOC + zipname

    if zipfile.exists():
        _log.info("%s already exists", zipfile)
        return

    _log.info("downloading data set %s", name)
    with zipfile.open("wb") as zf:
        res = urlopen(zipurl)
        block = res.read(8 * 1024 * 1024)
        while len(block):
            _log.debug("received %d bytes", len(block))
            zf.write(block)
            block = res.read(8 * 1024 * 1024)


def _fetch_main():
    logging.basicConfig(stream=sys.stderr, level=logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument("name", nargs="+", help="the name(s) of the dataset to fetch")
    parser.add_argument(
        "--data-dir", metavar="DIR", help="save extracted data to DIR", default="data"
    )
    args = parser.parse_args()

    dir = Path(args.data_dir)
    _log.info("extracting data to %s", dir)
    for name in args.name:
        _log.info("fetching data set %s", name)
        if name.startswith("ml-"):
            fetch_ml(name, dir)
        else:
            _log.error("unknown data set %s", name)
            raise ValueError("invalid data set")


if __name__ == "__main__":
    _fetch_main()



"""
PyTorch utility functions.
"""

import functools

import torch


def inference_mode(func):
    """
    Function decorator that puts PyTorch in inference mode.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with torch.inference_mode():
            return func(*args, **kwargs)

    return wrapper



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Generator
from typing import Literal, TypeAlias

import numpy as np
import structlog
import torch
from pydantic import BaseModel
from typing_extensions import NamedTuple, override

from lenskit.data import Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.data.types import UIPair
from lenskit.parallel.config import ensure_parallel_init
from lenskit.pipeline import Component
from lenskit.training import IterativeTraining, TrainingOptions

EntityClass: TypeAlias = Literal["user", "item"]


class ALSConfig(BaseModel):
    """
    Configuration for ALS scorers.
    """

    features: int = 50
    """
    The numer of latent features to learn.
    """
    epochs: int = 10
    """
    The number of epochs to train.
    """
    regularization: float | UIPair[float] = 0.1
    """
    L2 regularization strength.
    """
    save_user_features: bool = True
    """
    Whether to retain user feature values after training.
    """

    @property
    def user_reg(self) -> float:
        if isinstance(self.regularization, UIPair):
            return self.regularization.user
        else:
            return self.regularization

    @property
    def item_reg(self) -> float:
        if isinstance(self.regularization, UIPair):
            return self.regularization.item
        else:
            return self.regularization


class TrainContext(NamedTuple):
    """
    Context object for one half of an ALS training operation.
    """

    label: str
    matrix: torch.Tensor
    left: torch.Tensor
    right: torch.Tensor
    reg: float
    nrows: int
    ncols: int
    embed_size: int
    regI: torch.Tensor

    @classmethod
    def create(
        cls, label: str, matrix: torch.Tensor, left: torch.Tensor, right: torch.Tensor, reg: float
    ) -> TrainContext:
        nrows, ncols = matrix.shape
        lnr, embed_size = left.shape
        assert lnr == nrows
        assert right.shape == (ncols, embed_size)
        regI = torch.eye(embed_size, dtype=left.dtype, device=left.device) * reg
        return TrainContext(label, matrix, left, right, reg, nrows, ncols, embed_size, regI)


class TrainingData(NamedTuple):
    """
    Data for training the ALS model.
    """

    users: Vocabulary
    "User ID mapping."
    items: Vocabulary
    "Item ID mapping."
    ui_rates: torch.Tensor
    "User-item rating matrix."
    iu_rates: torch.Tensor
    "Item-user rating matrix."

    @property
    def n_users(self):
        return len(self.users)

    @property
    def n_items(self):
        return len(self.items)

    @classmethod
    def create(cls, users: Vocabulary, items: Vocabulary, ratings: torch.Tensor) -> TrainingData:
        assert ratings.shape == (len(users), len(items))

        transposed = ratings.transpose(0, 1).to_sparse_csr()
        return cls(users, items, ratings, transposed)

    def to(self, device):
        """
        Move the training data to another device.
        """
        return self._replace(ui_rates=self.ui_rates.to(device), iu_rates=self.iu_rates.to(device))


class ALSBase(IterativeTraining, Component[ItemList], ABC):
    """
    Base class for ALS models.

    Stability:
        Caller
    """

    config: ALSConfig

    users_: Vocabulary | None
    items_: Vocabulary
    user_features_: torch.Tensor | None
    item_features_: torch.Tensor

    logger: structlog.stdlib.BoundLogger

    @override
    def training_loop(
        self, data: Dataset, options: TrainingOptions
    ) -> Generator[dict[str, float], None, None]:
        """
        Run ALS to train a model.

        Args:
            ratings: the ratings data frame.

        Returns:
            ``True`` if the model was trained.
        """
        ensure_parallel_init()

        rng = options.random_generator()

        train = self.prepare_data(data)
        self.users_ = train.users
        self.items_ = train.items

        self.initialize_params(train, rng)

        return self._training_loop_generator(train)

        for algo in self.fit_iters(data, options):
            pass  # we just need to do the iterations

        return True

    def _training_loop_generator(
        self, train: TrainingData
    ) -> Generator[dict[str, float], None, None]:
        """
        Run ALS to train a model, yielding after each iteration.

        Args:
            ratings: the ratings data frame.
        """
        log = self.logger = self.logger.bind(features=self.config.features)

        assert self.user_features_ is not None
        assert self.item_features_ is not None
        u_ctx = TrainContext.create(
            "user", train.ui_rates, self.user_features_, self.item_features_, self.config.user_reg
        )
        i_ctx = TrainContext.create(
            "item", train.iu_rates, self.item_features_, self.user_features_, self.config.item_reg
        )

        for epoch in range(self.config.epochs):
            log = log.bind(epoch=epoch)
            epoch = epoch + 1

            du = self.als_half_epoch(epoch, u_ctx)
            log.debug("finished user epoch")

            di = self.als_half_epoch(epoch, i_ctx)
            log.debug("finished item epoch")

            log.debug("finished epoch (|ΔP|=%.3f, |ΔQ|=%.3f)", du, di)
            yield {"deltaP": du, "deltaQ": di}

        if not self.config.save_user_features:
            self.user_features_ = None
            self.user_ = None

        log.debug("finalizing model training")
        self.finalize_training()

    @abstractmethod
    def prepare_data(self, data: Dataset) -> TrainingData:  # pragma: no cover
        """
        Prepare data for training this model.  This takes in the ratings, and is
        supposed to do two things:

        -   Normalize or transform the rating/interaction data, as needed, for
            training.
        -   Store any parameters learned from the normalization (e.g. means) in
            the appropriate member variables.
        -   Return the training data object to use for model training.
        """
        ...

    def initialize_params(self, data: TrainingData, rng: np.random.Generator):
        """
        Initialize the model parameters at the beginning of training.
        """
        self.logger.debug("initializing item matrix")
        self.item_features_ = self.initial_params(data.n_items, self.config.features, rng)
        self.logger.debug("|Q|: %f", torch.norm(self.item_features_, "fro"))

        self.logger.debug("initializing user matrix")
        self.user_features_ = self.initial_params(data.n_users, self.config.features, rng)
        self.logger.debug("|P|: %f", torch.norm(self.user_features_, "fro"))

    @abstractmethod
    def initial_params(
        self, nrows: int, ncols: int, rng: np.random.Generator
    ) -> torch.Tensor:  # pragma: no cover
        """
        Compute initial parameter values of the specified shape.
        """
        ...

    @abstractmethod
    def als_half_epoch(self, epoch: int, context: TrainContext) -> float:  # pragma: no cover
        """
        Run one half of an ALS training epoch.
        """
        ...

    def finalize_training(self):
        pass

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        user_id = query.user_id
        user_num = None
        if user_id is not None and self.users_ is not None:
            user_num = self.users_.number(user_id, missing=None)

        u_offset = None
        u_feat = None
        if query.user_items is not None and len(query.user_items) > 0:
            u_feat, u_offset = self.new_user_embedding(user_num, query.user_items)

        if u_feat is None:
            if user_num is None or self.user_features_ is None:
                return ItemList(items, scores=np.nan)
            u_feat = self.user_features_[user_num, :]

        item_nums = items.numbers("torch", vocabulary=self.items_, missing="negative")
        item_mask = item_nums >= 0
        i_feats = self.item_features_[item_nums[item_mask], :]

        scores = torch.full((len(items),), np.nan, dtype=torch.float64)
        scores[item_mask] = i_feats @ u_feat

        results = ItemList(items, scores=scores)
        return self.finalize_scores(user_num, results, u_offset)

    @abstractmethod
    def new_user_embedding(
        self, user_num: int | None, items: ItemList
    ) -> tuple[torch.Tensor, float | None]:  # pragma: no cover
        """
        Generate an embedding for a user given their current ratings.
        """
        ...

    def finalize_scores(
        self, user_num: int | None, items: ItemList, user_bias: float | None
    ) -> ItemList:
        """
        Perform any final transformation of scores prior to returning them.
        """
        return items

    def __getstate__(self):
        return {k: v for (k, v) in self.__dict__.items() if k != "logger"}



"""
Console and related logging support
"""

import atexit
from logging import Handler, LogRecord

from rich.ansi import AnsiDecoder
from rich.console import Console
from rich.live import Live

console = Console(stderr=True)
_live: Live | None = None


class ConsoleHandler(Handler):
    """
    Lightweight Rich log handler for routing StructLog-formatted logs.
    """

    _decoder = AnsiDecoder()

    @property
    def supports_color(self) -> bool:
        return console.is_terminal and not console.no_color

    def emit(self, record: LogRecord) -> None:
        try:
            fmt = self.format(record)
            print(fmt)
            # console.print(*self._decoder.decode(fmt))
        except Exception:
            self.handleError(record)


def get_live() -> Live | None:
    return _live


def setup_console():
    global _live
    if _live is not None:
        return
    if not console.is_terminal:
        return

    _live = Live(console=console, transient=True)
    _live.start()


@atexit.register
def _stop_console():
    if _live is not None:
        _live.stop()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit dataset abstraction.
"""

# pyright: basic
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    Any,
    Literal,
    TypeAlias,
    TypeVar,
    overload,
)

import numpy as np
import pandas as pd
import scipy.sparse as sps
import torch

from .items import ItemList
from .tables import NumpyUserItemTable, TorchUserItemTable
from .types import ID
from .vocab import Vocabulary

DF_FORMAT: TypeAlias = Literal["numpy", "pandas", "torch"]
MAT_FORMAT: TypeAlias = Literal["scipy", "torch", "pandas", "structure"]
MAT_AGG: TypeAlias = Literal["count", "sum", "mean", "first", "last"]
LAYOUT: TypeAlias = Literal["csr", "coo"]
ACTION_FIELDS: TypeAlias = Literal["ratings", "timestamps"] | str

K = TypeVar("K")


class FieldError(KeyError):
    """
    The requested field does not exist.
    """

    def __init__(self, entity, field):
        super().__init__(f"{entity}[{field}]")


class Dataset(ABC):
    """
    Representation of a data set for LensKit training, evaluation, etc. Data can
    be accessed in a variety of formats depending on the needs of a component.

    .. note::
        Zero-copy conversions are used whenever possible, so client code must not
        modify returned data in-place.

    .. todo::
        Support for advanced rating situations is not yet supported:

        * repeated ratings
        * mixed implicit & explicit feedback
        * later actions removing earlier ratings

    .. todo::
        Support for item and user content or metadata is not yet implemented.

    .. stability:: caller
    """

    _item_stats: pd.DataFrame | None = None
    _user_stats: pd.DataFrame | None = None

    @property
    @abstractmethod
    def items(self) -> Vocabulary:
        """
        The items known by this dataset.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def users(self) -> Vocabulary:
        """
        The users known by this dataset.
        """
        raise NotImplementedError()

    @abstractmethod
    def count(self, what: str) -> int:
        """
        Count entities in the dataset.

        .. note::

            The precise counts are subtle in the presence of repeated or
            superseded interactions. See :meth:`interaction_count` and
            :meth:`rating_count` for details on the ``"interactions"`` and
            ``"ratings"`` counts.

        Args:
            what:
                The type of entity to count.  Commonly-supported ones include:

                * users
                * items
                * pairs (observed user-item pairs)
                * interactions
                * ratings
        """
        raise NotImplementedError()

    @property
    def item_count(self) -> int:
        return self.count("items")

    @property
    def user_count(self) -> int:
        return self.count("users")

    @property
    def interaction_count(self) -> int:
        """
        Count the total number of interaction records.  Equivalent to
        ``count("interactions")``.

        .. note::
            If the interaction records themselves reprsent counts, such as the
            number of times a song was played, this returns the number of
            *records*, not the total number of plays.
        """
        return self.count("interactions")

    @property
    def rating_count(self) -> int:
        """
        Count the total number of ratings (excluding superseded ratings).
        Equivalent to ``count("ratings")``.
        """
        return self.count("ratings")

    @overload
    @abstractmethod
    def interaction_log(
        self,
        format: Literal["pandas"],
        *,
        fields: str | list[str] | None = "all",
        original_ids: bool = False,
    ) -> pd.DataFrame: ...
    @overload
    @abstractmethod
    def interaction_log(
        self, format: Literal["numpy"], *, fields: str | list[str] | None = "all"
    ) -> NumpyUserItemTable: ...
    @overload
    @abstractmethod
    def interaction_log(
        self, format: Literal["torch"], *, fields: str | list[str] | None = "all"
    ) -> TorchUserItemTable: ...
    @abstractmethod
    def interaction_log(
        self,
        format: str,
        *,
        fields: str | list[str] | None = "all",
        original_ids: bool = False,
    ) -> Any:
        """
        Get the user-item interactions as a table in the requested format. The
        table is not in a specified order.  Interactions may be repeated (e.g.
        the same user may listen to a song multiple times).  For a non-repeated
        “ratings matrix” view of the data, see :meth:`interaction_matrix`.

        .. warning::
            Client code **must not** perform in-place modifications on the table
            returned from this method.  Whenever possible, it will be a shallow
            view on top of the underlying storage, and modifications may corrupt
            data for other code.

        Args:
            format:
                The desired data format.  Currently-supported formats are:

                * ``"pandas"`` — returns a :class:`pandas.DataFrame`.  The index
                  is not meaningful.
                * ``"numpy"`` — returns a :class:`~tables.NumpyUserItemTable`.
                * ``"torch"`` — returns a :class:`~tables.TorchUserItemTable`.
            fields:
                Which fields to include.  If set to ``"all"``, will include all
                available fields in the resulting table; ``None`` includes no
                fields besides the user and item.  Commonly-available fields
                include ``"rating"`` and ``"timestamp"``.  Missing fields will
                be omitted in the result.
            original_ids:
                If ``True``, return user and item IDs as represented in the
                original source data in columns named ``user_id`` and
                ``item_id``, instead of the user and item numbers typically
                returned.  Only applicable to the ``pandas`` format. See
                :ref:`data-identifiers`.

        Returns:
            The user-item interaction log in the specified format.
        """
        raise NotImplementedError()

    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        format: Literal["pandas"],
        *,
        layout: Literal["coo"] | None = None,
        field: str | None = None,
        combine: MAT_AGG | None = None,
        original_ids: bool = False,
    ) -> pd.DataFrame: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        format: Literal["torch"],
        *,
        layout: Literal["csr", "coo"] | None = None,
        field: str | None = None,
        combine: MAT_AGG | None = None,
    ) -> torch.Tensor: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        format: Literal["scipy"],
        *,
        layout: Literal["coo"],
        legacy: Literal[True],
        field: str | None = None,
        combine: MAT_AGG | None = None,
    ) -> sps.coo_matrix: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        format: Literal["scipy"],
        *,
        layout: Literal["coo"],
        legacy: bool = False,
        field: str | None = None,
        combine: MAT_AGG | None = None,
    ) -> sps.coo_array: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        format: Literal["scipy"],
        *,
        layout: Literal["csr"] | None = None,
        legacy: Literal[True],
        field: str | None = None,
        combine: MAT_AGG | None = None,
    ) -> sps.csr_matrix: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        format: Literal["scipy"],
        *,
        layout: Literal["csr"] | None = None,
        legacy: bool = False,
        field: str | None = None,
        combine: MAT_AGG | None = None,
    ) -> sps.csr_array: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        format: Literal["structure"],
        *,
        layout: Literal["csr"] | None = None,
    ) -> CSRStructure: ...
    @abstractmethod
    def interaction_matrix(
        self,
        format: str,
        *,
        layout: str | None = None,
        legacy: bool = False,
        field: str | None = None,
        combine: str | None = None,
        original_ids: bool = False,
    ) -> Any:
        """
        Get the user-item interactions as “ratings” matrix.  Interactions are
        not repeated.  The matrix may be in “coordinate” format, in which case
        it is comparable to :meth:`interaction_log` but without repeated
        interactions, or it may be in a compressed sparse format.

        .. todo::
            Aggregate is currently ignored because repeated interactions are not
            yet supported.

        .. warning::
            Client code **must not** perform in-place modifications on the matrix
            returned from this method.  Whenever possible, it will be a shallow
            view on top of the underlying storage, and modifications may corrupt
            data for other code.

        Args:
            format:
                The desired data format.  Currently-supported formats are:

                * ``"pandas"`` — returns a :class:`pandas.DataFrame`.
                * ``"torch"`` — returns a sparse :class:`torch.Tensor` (see
                  :mod:`torch.sparse`).
                * ``"scipy"`` — returns a sparse array from :mod:`scipy.sparse`.
                * ``"structure"`` — returns a :class:`~matrix.CSRStructure`
                  containing only the user and item numbers in compressed sparse
                  row format.
            field:
                Which field to return in the matrix.  Common fields include
                ``"rating"`` and ``"timestamp"``.

                If unspecified (``None``), this will yield an implicit-feedback
                indicator matrix, with 1s for observed items; the ``"pandas"``
                format will only include user and item columns.

                If the ``rating`` field is requested but is not defined in the
                underlying data, then this is equivalent to ``"indicator"``,
                except that the ``"pandas"`` format will include a ``"rating"``
                column of all 1s.

                The ``"pandas"`` format also supports the special field name
                ``"all"`` to return a data frame with all available fields. When
                ``field="all"``, a field named ``count`` (if defined) is
                combined with the ``sum`` method, and other fields use ``last``.
            combine:
                How to combine multiple observations for a single user-item
                pair. Available methods are:

                * ``"count"`` — count the user-item interactions. Only valid
                  when ``field=None``; if the underlying data defines a
                  ``count`` field, then this is equivalent to ``"sum"`` on that
                  field.
                * ``"sum"`` — sum the field values.
                * ``"first"``, ``"last"`` — take the first or last value seen
                  (in timestamp order, if timestamps are defined; otherwise,
                  their order in the original input).
            layout:
                The layout for a sparse matrix.  Can be either ``csr`` or
                ``coo``, or ``None`` to use the default for the specified
                format.  CSR is only supported by Torch and SciPy backends.
            legacy:
                ``True`` to return a legacy SciPy sparse matrix instead of
                sparse array.
            original_ids:
                ``True`` to return user and item IDs instead of numbers in
                ``pandas``-format matrix.
        """
        raise NotImplementedError()

    @abstractmethod
    @overload
    def user_row(self, user_id: ID) -> ItemList | None: ...
    @abstractmethod
    @overload
    def user_row(self, *, user_num: int) -> ItemList: ...
    @abstractmethod
    def user_row(
        self, user_id: ID | None = None, *, user_num: int | None = None
    ) -> ItemList | None:
        """
        Get a user's row from the interaction matrix.  Available fields are
        returned as fields. If the dataset has ratings, these are provided as a
        ``rating`` field, **not** as the item scores.  The item list is
        unordered, but items are returned in order by item number.

        Args:
            user_id:
                The ID of the user to retrieve.
            user_num:
                The number of the user to retrieve.

        Returns:
            The user's interaction matrix row, or ``None`` if no user with that ID
            exists.
        """
        raise NotImplementedError()

    def item_stats(self) -> pd.DataFrame:
        """
        Get item statistics.

        Returns:
            A data frame indexed by item ID with the following columns:

            * count — the number of interactions recorded for this item.
            * user_count — the number of distinct users who have interacted with
              or rated this item.
            * rating_count — the number of ratings for this item.  Only provided
              if the dataset has explicit ratings; if there are repeated
              ratings, this does **not** count superseded ratings.
            * mean_rating — the mean of the ratings. Only provided if the
              dataset has explicit ratings.
            * first_time — the first time the item appears. Only provided if the
              dataset has timestamps.

            The index is the vocabulary, so ``iloc`` works with item numbers.
        """

        if self._item_stats is None:
            log = self.interaction_log("numpy")

            counts = np.zeros(self.item_count, dtype=np.int32)
            np.add.at(counts, log.item_nums, 1)
            frame = pd.DataFrame(
                {
                    "count": counts,
                    "user_count": counts,
                },
                index=self.items.index,
            )

            if log.ratings is not None:
                sums = np.zeros(self.item_count, dtype=np.float64)
                np.add.at(sums, log.item_nums, log.ratings)
                frame["rating_count"] = counts
                frame["mean_rating"] = sums / counts

            if log.timestamps is not None:
                i64i = np.iinfo(np.int64)
                times = np.full(self.item_count, i64i.max, dtype=np.int64)
                np.minimum.at(times, log.item_nums, log.timestamps)
                frame["first_time"] = times

            self._item_stats = frame

        return self._item_stats

    def user_stats(self) -> pd.DataFrame:
        """
        Get user statistics.

        Returns:
            A data frame indexed by user ID with the following columns:

            * count — the number of interactions recorded for this user.
            * item_count — the number of distinct items with which this user has
              interacted.
            * rating_count — the number of ratings for this user.  Only provided
              if the dataset has explicit ratings; if there are repeated
              ratings, this does **not** count superseded ratings.
            * mean_rating — the mean of the user's reatings. Only provided if
              the dataset has explicit ratings.
            * first_time — the first time the user appears. Only provided if the
              dataset has timestamps.
            * last_time — the last time the user appears. Only provided if the
              dataset has timestamps.

            The index is the vocabulary, so ``iloc`` works with user numbers.
        """

        if self._user_stats is None:
            log = self.interaction_log("numpy")

            counts = np.zeros(self.user_count, dtype=np.int32)
            np.add.at(counts, log.user_nums, 1)
            frame = pd.DataFrame(
                {
                    "count": counts,
                    "item_count": counts,
                },
                index=self.users.index,
            )

            if log.ratings is not None:
                sums = np.zeros(self.user_count, dtype=np.float64)
                np.add.at(sums, log.user_nums, log.ratings)
                frame["rating_count"] = counts
                frame["mean_rating"] = sums / counts

            if log.timestamps is not None:
                i64i = np.iinfo(np.int64)
                first = np.full(self.user_count, i64i.max, dtype=np.int64)
                last = np.full(self.user_count, i64i.min, dtype=np.int64)
                np.minimum.at(first, log.user_nums, log.timestamps)
                np.maximum.at(last, log.user_nums, log.timestamps)
                frame["first_time"] = first
                frame["last_time"] = last

            self._user_stats = frame

        return self._user_stats


from .matrix import CSRStructure  # noqa: E402



# pyright: strict
from __future__ import annotations

from threading import Lock
from uuid import UUID, uuid4

import structlog
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)
from rich.progress import Progress as ProgressImpl
from rich.text import Text
from typing_extensions import override

from .._console import console, get_live
from .._proxy import get_logger
from ._base import Progress

_log = get_logger("lenskit.logging.progress")
_pb_lock = Lock()
_progress: ProgressImpl | None = None
_active_bars: dict[UUID, RichProgress] = {}


class RichProgress(Progress):
    """
    Progress bark backed by Rich.
    """

    uuid: UUID
    label: str
    total: int | None
    fields: dict[str, str | None]
    logger: structlog.stdlib.BoundLogger
    _task: TaskID | None = None

    def __init__(self, label: str, total: int | None, fields: dict[str, str | None]):
        super().__init__()
        self.uuid = uuid4()
        self.label = label
        self.total = total
        self.fields = fields

        self.logger = _log.bind(label=label, uuid=str(self.uuid))

        self._task = _install_bar(self)

    def update(self, advance: int = 1, **kwargs: float | int | str):
        if _progress is not None:
            _progress.update(self._task, advance=advance, **kwargs)  # type: ignore

    def finish(self):
        _remove_bar(self)


def _install_bar(bar: RichProgress) -> TaskID | None:
    global _progress
    bar.logger.debug("installing progress bar")
    live = get_live()
    if live is None:
        return None

    with _pb_lock:
        if _progress is None:
            _progress = ProgressImpl(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                BarColumn(),
                MofNCompleteColumn(),
                RateColumn(),
                TaskProgressColumn(),
                TimeRemainingColumn(),
                console=console,
            )
            live.update(_progress)

        _active_bars[bar.uuid] = bar
        return _progress.add_task(bar.label, total=bar.total)


def _remove_bar(bar: RichProgress):
    live = get_live()
    if live is None or _progress is None:
        return
    if bar.uuid not in _active_bars:
        return
    if bar._task is None:
        return

    bar.logger.debug("uninstalling progress bar")

    with _pb_lock:
        _progress.remove_task(bar._task)
        del _active_bars[bar.uuid]


class RateColumn(ProgressColumn):
    def __init__(self):
        super().__init__()

    @override
    def render(self, task: Task):
        speed = task.finished_speed or task.speed
        if speed is None:
            disp = "?"
        elif speed > 1000:
            disp = "{:d} it/s".format(int(speed))
        elif speed > 1:
            disp = "{:.3g} it/s".format(speed)
        else:
            disp = "{:.3g} s/it".format(1.0 / speed)

        return Text(disp, "progress.percentage")



"""
MovieLens test fixtures and data marks.
"""

import os
from typing import Generator

import numpy as np
import pandas as pd
from pyprojroot import here

import pytest

from lenskit.basic import PopScorer, SoftmaxRanker
from lenskit.batch import recommend
from lenskit.data import Dataset, ItemListCollection, UserIDKey, from_interactions_df
from lenskit.data.lazy import LazyDataset
from lenskit.data.movielens import load_movielens, load_movielens_df
from lenskit.logging import get_logger
from lenskit.pipeline import RecPipelineBuilder
from lenskit.splitting import TTSplit, simple_test_pair

_log = get_logger("lenskit.testing")

ml_test_dir = here("data/ml-latest-small")
ml_100k_zip = here("data/ml-100k.zip")

ml_test: Dataset = LazyDataset(lambda: load_movielens(ml_test_dir))

retrain = os.environ.get("LK_TEST_RETRAIN")


@pytest.fixture(scope="session")
def ml_ratings() -> Generator[pd.DataFrame, None, None]:
    """
    Fixture to load the test MovieLens ratings as a data frame. To use this,
    just include it as a parameter in your test::

        def test_thing_with_data(ml_ratings: pd.DataFrame):
            ...

    .. note::
        This is imported in ``conftest.py`` so it is always available in LensKit tests.
    """
    yield load_movielens_df(ml_test_dir)


@pytest.fixture(scope="session")
def ml_ds_unchecked(ml_ratings: pd.DataFrame) -> Generator[Dataset, None, None]:
    """
    Fixture to load the MovieLens dataset, without checking for modifications.

    Usually use :func:`ml_ds` instead.
    """
    yield from_interactions_df(ml_ratings)


@pytest.fixture(scope="function" if retrain else "module")
def ml_ds(ml_ds_unchecked: Dataset) -> Generator[Dataset, None, None]:
    """
    Fixture to load the MovieLens test dataset.  To use this, just include it as
    a parameter in your test::

        def test_thing_with_data(ml_ds: Dataset):
            ...

    .. note::
        This is imported in ``conftest.py`` so it is always available in LensKit tests.
    """
    log = _log.bind()

    ds = ml_ds_unchecked
    old_rates = ds.interaction_matrix("pandas", field="rating", original_ids=True).copy(deep=True)
    old_ustats = ds.user_stats().copy(deep=True)
    old_istats = ds.item_stats().copy(deep=True)

    yield ds

    ustats = ds.user_stats()
    istats = ds.item_stats()

    rates = ds.interaction_matrix("pandas", field="rating", original_ids=True)
    assert rates["rating"].values == pytest.approx(old_rates["rating"].values)

    for col in old_ustats.columns:
        log.info("checking user stats column", column=col)
        assert ustats[col].values == pytest.approx(old_ustats[col].values)

    for col in old_istats.columns:
        log.info("checking item stats column", column=col)
        assert istats[col].values == pytest.approx(old_istats[col].values)


@pytest.fixture
def ml_100k() -> Generator[pd.DataFrame, None, None]:
    """
    Fixture to load the MovieLens 100K dataset (currently as a data frame).  It skips
    the test if the ML100K data is not available.
    """
    if not ml_100k_zip.exists():
        pytest.skip("ML100K data not available")
    yield load_movielens_df(ml_100k_zip)


@pytest.fixture(scope="session")
def demo_recs() -> tuple[TTSplit, ItemListCollection[UserIDKey]]:
    """
    A demo set of train, test, and recommendation data.
    """
    rng = np.random.default_rng(42)
    ml_ds = load_movielens(ml_test_dir)
    split = simple_test_pair(ml_ds, f_rates=0.5, rng=rng)

    builder = RecPipelineBuilder()
    builder.scorer(PopScorer())
    builder.ranker(SoftmaxRanker(n=500))
    pipe = builder.build()
    pipe.train(split.train)

    recs = recommend(pipe, list(split.test.keys()), 500, n_jobs=1, rng=rng)
    return split, recs



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
User-based k-NN collaborative filtering.
"""

# pyright: basic
from __future__ import annotations

import warnings

import numpy as np
import pandas as pd
import structlog
import torch
from pydantic import BaseModel, PositiveFloat, PositiveInt, field_validator
from scipy.sparse import csc_array
from typing_extensions import NamedTuple, Optional, override

from lenskit import util
from lenskit.data import Dataset, FeedbackType, ItemList, QueryInput, RecQuery
from lenskit.data.vocab import Vocabulary
from lenskit.diagnostics import DataWarning
from lenskit.logging import get_logger
from lenskit.math.sparse import normalize_sparse_rows, safe_spmv, torch_sparse_to_scipy
from lenskit.parallel.config import ensure_parallel_init
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_log = get_logger(__name__)


class UserKNNConfig(BaseModel, extra="forbid"):
    "Configuration for :class:`ItemKNNScorer`."

    k: PositiveInt = 20
    """
    The maximum number of neighbors for scoring each item.
    """
    min_nbrs: PositiveInt = 1
    """
    The minimum number of neighbors for scoring each item.
    """
    min_sim: PositiveFloat = 1.0e-6
    """
    Minimum similarity threshold for considering a neighbor.  Must be positive;
    if less than the smallest 32-bit normal (:math:`1.175 \\times 10^{-38}`), is
    clamped to that value.
    """
    feedback: FeedbackType = "explicit"
    """
    The type of input data to use (explicit or implicit).  This affects data
    pre-processing and aggregation.
    """

    @field_validator("min_sim", mode="after")
    @staticmethod
    def clamp_min_sim(sim) -> float:
        return max(sim, float(np.finfo(np.float64).smallest_normal))

    @property
    def explicit(self) -> bool:
        """
        Query whether this is in explicit-feedback mode.
        """
        return self.feedback == "explicit"


class UserKNNScorer(Component[ItemList], Trainable):
    """
    User-user nearest-neighbor collaborative filtering with ratings. This
    user-user implementation is not terribly configurable; it hard-codes design
    decisions found to work well in the previous Java-based LensKit code.

    .. note::

        This component must be used with queries containing the user's history,
        either directly in the input or by wiring its query input to the output of a
        user history component (e.g., :class:`~lenskit.basic.UserTrainingHistoryLookup`).

    Stability:
        Caller
    """

    config: UserKNNConfig

    users_: Vocabulary
    "The index of user IDs."
    items_: Vocabulary
    "The index of item IDs."
    user_means_: torch.Tensor | None
    "Mean rating for each known user."
    user_vectors_: torch.Tensor
    "Normalized rating matrix (CSR) to find neighbors at prediction time."
    user_ratings_: csc_array
    "Centered but un-normalized rating matrix (COO) to find neighbor ratings."

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        """
        "Train" a user-user CF model.  This memorizes the rating data in a format that is usable
        for future computations.

        Args:
            ratings(pandas.DataFrame): (user, item, rating) data for collaborative filtering.
        """
        if hasattr(self, "user_ratings_") and not options.retrain:
            return

        ensure_parallel_init()
        rmat = data.interaction_matrix("torch", field="rating" if self.config.explicit else None)
        assert rmat.is_sparse_csr

        if self.config.explicit:
            rmat, means = normalize_sparse_rows(rmat, "center")
            if np.allclose(rmat.values(), 0.0):
                _log.warning("normalized ratings are zero, centering is not recommended")
                warnings.warn(
                    "Ratings seem to have the same value, centering is not recommended.",
                    DataWarning,
                )
        else:
            means = None

        normed, _norms = normalize_sparse_rows(rmat, "unit")

        self.user_vectors_ = normed
        self.user_ratings_ = torch_sparse_to_scipy(rmat).tocsc()
        self.users_ = data.users.copy()
        self.user_means_ = means
        self.items_ = data.items.copy()

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        """
        Compute predictions for a user and items.

        Args:
            user: the user ID
            items (array-like): the items to predict
            ratings (pandas.Series):
                the user's ratings (indexed by item id); if provided, will be used to
                recompute the user's bias at prediction time.

        Returns:
            pandas.Series: scores for the items, indexed by item id.
        """
        query = RecQuery.create(query)
        watch = util.Stopwatch()
        log = _log.bind(user_id=query.user_id, n_items=len(items))
        if len(items) == 0:
            log.debug("no candidate items, skipping")
            return ItemList(items, scores=np.nan)

        udata = self._get_user_data(query)
        if udata is None:
            log.debug("user has no ratings, skipping")
            return ItemList(items, scores=np.nan)

        uidx, ratings, umean = udata
        assert ratings.shape == (len(self.items_),)  # ratings is a dense vector

        # now ratings has vbeen normalized to be a mean-centered unit vector
        # this means we can dot product to score neighbors
        # score the neighbors!
        nbr_sims = safe_spmv(self.user_vectors_, ratings)
        assert nbr_sims.shape == (len(self.users_),)
        if uidx is not None:
            # zero out the self-similarity
            nbr_sims[uidx] = 0

        # get indices for these neighbors
        nbr_idxs = torch.arange(len(self.users_), dtype=torch.int64)

        nbr_mask = nbr_sims >= self.config.min_sim

        kn_sims = nbr_sims[nbr_mask]
        kn_idxs = nbr_idxs[nbr_mask]
        if len(kn_sims) > 0:
            log.debug(
                "found %d candidate neighbors (of %d total), max sim %0.4f",
                len(kn_sims),
                len(self.users_),
                torch.max(kn_sims).item(),
            )
        else:
            log.warning("no candidate neighbors found")
            return ItemList(items, scores=np.nan)

        assert not torch.any(torch.isnan(kn_sims))

        iidxs = items.numbers(vocabulary=self.items_, missing="negative")
        iidxs = torch.from_numpy(iidxs).to(torch.int64)

        ki_mask = iidxs >= 0
        usable_iidxs = iidxs[ki_mask]

        scores = score_items_with_neighbors(
            log,
            usable_iidxs,
            kn_idxs,
            kn_sims,
            self.user_ratings_,
            self.config.k,
            self.config.min_nbrs,
            self.config.explicit,
        )

        scores += umean

        results = pd.Series(scores, index=items.ids()[ki_mask.numpy()], name="prediction")
        results = results.reindex(items.ids())

        log.debug(
            "scored %d items in %s",
            results.notna().sum(),
            watch,
        )
        return ItemList(items, scores=results.values)  # type: ignore

    def _get_user_data(self, query: RecQuery) -> Optional[UserRatings]:
        "Get a user's data for user-user CF"

        index = self.users_.number(query.user_id, missing=None)

        if query.user_items is None:
            if index is None:
                _log.warning("user %s has no ratings and none provided", query.user_id)
                return None

            assert index >= 0
            row = self.user_vectors_[index].to_dense()
            if self.config.explicit:
                assert self.user_means_ is not None
                umean = self.user_means_[index].item()
            else:
                umean = 0
            return UserRatings(index, row, umean)
        else:
            _log.debug("using provided item history")
            ratings = torch.zeros(len(self.items_), dtype=torch.float32)
            ui_nos = query.user_items.numbers("torch", missing="negative", vocabulary=self.items_)
            ui_mask = ui_nos >= 0

            if self.config.explicit:
                urv = query.user_items.field("rating", "torch")
                if urv is None:
                    _log.warning("user %s has items but no ratings", query.user_id)
                    return None

                umean = urv.mean().item()
                ratings[ui_nos[ui_mask]] = urv[ui_mask] - umean
            else:
                umean = 0
                ratings[ui_nos[ui_mask]] = 1.0

            return UserRatings(index, ratings, umean)


class UserRatings(NamedTuple):
    """
    Dense user ratings.
    """

    index: int | None
    ratings: torch.Tensor
    mean: float


def score_items_with_neighbors(
    log: structlog.stdlib.BoundLogger,
    items: torch.Tensor,
    nbr_rows: torch.Tensor,
    nbr_sims: torch.Tensor,
    ratings: csc_array,
    max_nbrs: int,
    min_nbrs: int,
    average: bool,
) -> np.ndarray[tuple[int], np.dtype[np.float64]]:
    # select a sub-matrix for further manipulation
    items = items.numpy()
    (ni,) = items.shape
    (nrow, ncol) = ratings.shape

    # sort neighbors by similarity
    nbr_order = np.argsort(-nbr_sims)
    nbr_rows = nbr_rows[nbr_order].numpy()
    nbr_sims = nbr_sims[nbr_order].numpy()

    # get the rating rows for our neighbors
    nbr_rates = ratings[nbr_rows, :]

    # which items are scorable?
    counts = np.diff(nbr_rates.indptr)
    min_nbr_mask = counts >= min_nbrs
    is_nbr_mask = min_nbr_mask[items]
    is_scorable = items[is_nbr_mask]

    # get the ratings for requested scorable items
    nbr_rates = nbr_rates[:, is_scorable]
    assert isinstance(nbr_rates, csc_array)
    nbr_rates.sort_indices()
    counts = counts[is_scorable]

    log.debug(
        "scoring items",
        max_count=np.max(counts) if len(counts) else 0,
        nbr_shape=nbr_rates.shape,
    )

    # Now, for our next trick - we have a CSC matrix, whose rows (users) are
    # sorted by decreasing similarity.  So we can *zero* any entries past the
    # first max_neighbors in a row.  This can be done with a little bit of
    # jiggery-pokery.

    # step 1: create a list of column start indices
    starts = np.repeat(nbr_rates.indptr[:-1], counts)
    # step 2: create a ranking from start to end
    ranks = np.arange(nbr_rates.nnz, dtype=np.int32)
    # step 3: subtract the column starts — this will give us numbers within rows
    ranks -= starts
    rmask = ranks >= max_nbrs
    # step 4: zero out rating values for everything past max_nbrs
    nbr_rates.data[rmask] = 0

    # now we can just do a matrix-vector multiply to compute the scores
    results = np.full(ni, np.nan)
    results[is_nbr_mask] = nbr_rates.T @ nbr_sims

    if average:
        nbr_ones = csc_array(
            (np.where(rmask, 0, 1), nbr_rates.indices, nbr_rates.indptr), nbr_rates.shape
        )
        tot_sims = nbr_ones.T @ nbr_sims
        assert np.all(np.isfinite(tot_sims))
        results[is_nbr_mask] /= tot_sims

    return results



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
"""
Interfaces and support for model training.
"""

# pyright: strict
from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass
from typing import Protocol, runtime_checkable

import numpy as np

from lenskit.data.dataset import Dataset
from lenskit.logging import get_logger, item_progress
from lenskit.random import RNGInput, random_generator

_log = get_logger(__name__)


@dataclass(frozen=True)
class TrainingOptions:
    """
    Options and context settings that govern model training.
    """

    retrain: bool = True
    """
    Whether the model should retrain if it is already trained.  If ``False``,
    the model should cleanly skip training if it is already trained.
    """

    device: str | None = None
    """
    The device on which to train (e.g. ``'cuda'``).  May be ignored if the model
    does not support the specified device.
    """

    rng: RNGInput = None
    """
    Random number generator to use for any randomness in the training process.
    This option contains any `SPEC 7`_-compatible random number generator
    specification; the :func:`~lenskit.random.random_generator` will convert
    that into a NumPy :class:`~numpy.random.Generator`.
    """

    def random_generator(self) -> np.random.Generator:
        """
        Obtain a random generator from the configured RNG or seed.

        .. note::

            Each call to this method will return a fresh generator from the same
            seed.  Components should call it once at the beginning of their
            training procesess.
        """
        return random_generator(self.rng)


@runtime_checkable
class Trainable(Protocol):  # pragma: nocover
    """
    Interface for components and objects that can learn parameters from training
    data. It supports training and checking if a component has already been
    trained.  The resulting model should be pickleable.  Trainable objects are
    usually also components.

    .. note::

        Trainable components must also implement ``__call__``.

    .. note::

        A future LensKit version will add support for extracting model
        parameters a la Pytorch's ``state_dict``, but this capability was not
        ready for 2025.1.

    Stability:
        Full
    """

    def train(self, data: Dataset, options: TrainingOptions) -> None:
        """
        Train the model to learn its parameters from a training dataset.

        Args:
            data:
                The training dataset.
            options:
                The training options.
        """
        raise NotImplementedError()


class IterativeTraining(ABC, Trainable):
    """
    Base class for components that support iterative training.  This both
    automates the :meth:`Trainable.train` method for iterative training in terms
    of initialization, epoch, and finalization methods, and exposes those
    methods to client code that may wish to directly control the iterative
    training process.

    Stability:
        Full
    """

    trained_epochs: int = 0
    """
    The number of epochs for which this model has been trained.
    """

    @property
    def expected_training_epochs(self) -> int | None:
        """
        Get the number of training epochs expected to run.  The default
        implementation looks for an ``epochs`` attribute on the configuration
        object (``self.config``).
        """
        cfg = getattr(self, "config", None)
        if cfg:
            return getattr(cfg, "epochs", None)

    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()) -> None:
        """
        Implementation of :meth:`Trainable.train` that uses the training loop.
        It also uses the :attr:`trained_epochs` attribute to detect if the model
        has already been trained for the purposes of honoring
        :attr:`TrainingOptions.retrain`, and updates that attribute as model
        training progresses.
        """
        if self.trained_epochs > 0 and not options.retrain:
            return

        self.trained_epochs = 0
        log = _log.bind(model=f"{self.__class__.__module__}.{self.__class__.__qualname__}")
        log.info("training model")
        n = self.expected_training_epochs
        log.debug("creating training loop")
        loop = self.training_loop(data, options)
        log.debug("beginning training iterations")
        with item_progress("Training iterations", total=n) as pb:
            for i, metrics in enumerate(loop, 1):
                metrics = metrics or {}
                log.info("finished epoch", epoch=i, **metrics)
                self.trained_epochs += 1
                pb.update()

        log.info("model training finished", epochs=self.trained_epochs)

    @abstractmethod
    def training_loop(
        self, data: Dataset, options: TrainingOptions
    ) -> Iterator[dict[str, float] | None]:
        """
        Training loop implementation, to be supplied by the derived class.  This
        method should return a iterator that, when iterated, will perform each
        training epoch; when training is complete, it should finalize the model
        and signal iteration completion.

        Each epoch can yield metrics, such as training or validation loss, to be
        logged with structured logging and can be used by calling code to do
        other analysis.

        See :ref:`iterative-training` for more details on writing iterative
        training loops.
        """
        raise NotImplementedError()



"""
Components that look up user history from the training data.
"""

from __future__ import annotations

import logging
from typing import Literal

import numpy as np
import pandas as pd
from scipy.sparse import csr_array
from typing_extensions import override

from lenskit.data import Dataset, ItemList, QueryInput, RecQuery
from lenskit.data.matrix import CSRStructure
from lenskit.data.vocab import Vocabulary
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_logger = logging.getLogger(__name__)


class UserTrainingHistoryLookup(Component[ItemList], Trainable):
    """
    Look up a user's history from the training data.

    Stability:
        Caller
    """

    training_data_: Dataset

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        # TODO: find a better data structure for this
        if hasattr(self, "training_data_") and not options.retrain:
            return

        self.training_data_ = data

    def __call__(self, query: QueryInput) -> RecQuery:
        """
        Look up the user's data from the training history (if needed), and
        ensure a fully-populated :class:`RecQuery`.
        """
        query = RecQuery.create(query)
        if query.user_id is None:
            return query

        if query.user_items is None:
            query.user_items = self.training_data_.user_row(query.user_id)

        return query


class KnownRatingScorer(Component[ItemList], Trainable):
    """
    Score items by returning their values from the training data.

    Stability:
        Caller

    Args:
        score:
            Whether to score items with their rating values, or a 0/1 indicator
            of their presence in the training data.  The default (``None``) uses
            ratings if available, and otherwise scores with 1 for interacted
            items and leaves non-interacted items unscored.
        source:
            Whether to use the training data or the user's history represented
            in the query as the source of score data.
    """

    score: Literal["rating", "indicator"] | None
    source: Literal["training", "query"]

    users_: Vocabulary
    items_: Vocabulary
    matrix_ = csr_array | CSRStructure

    def __init__(
        self,
        score: Literal["rating", "indicator"] | None = None,
        source: Literal["training", "query"] = "training",
    ):
        self.score = score
        self.source = source

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "matrix_") and not options.retrain:
            return

        if self.source == "query":
            return

        self.users_ = data.users
        self.items_ = data.items
        if self.score == "indicator":
            self.matrix_ = data.interaction_matrix("structure")
        else:
            self.matrix_ = data.interaction_matrix("scipy", field="rating")

    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        # figure out what scores we start with
        scores = None
        if self.source == "query" and query.user_items is not None:
            if self.score != "indicator":
                scores = query.user_items.field("rating", "pandas", index="ids")
            if scores is None:
                scores = pd.Series(1.0, index=query.user_items.ids())

        elif (
            self.source == "training" and query.user_id is not None and query.user_id in self.users_
        ):
            urow = self.users_.number(query.user_id)
            if isinstance(self.matrix_, csr_array):
                assert self.score != "indicator"
                # get the user's row as a sparse array
                uarr = self.matrix_[[urow]]
                assert isinstance(uarr, csr_array)
                # create a series
                scores = pd.Series(uarr.data, index=self.items_.ids(uarr.indices))
            elif isinstance(self.matrix_, CSRStructure):
                scores = pd.Series(1.0, index=self.items_.ids(self.matrix_.row_cs(urow)))

        if scores is None:
            scores = pd.Series(np.nan, index=items.ids())

        scores = scores.reindex(
            items.ids(), fill_value=0.0 if self.score == "indicator" else np.nan
        )
        return ItemList(items, scores=scores.values)  # type: ignore



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Batch-run recommendation pipelines for evaluation.
"""

from __future__ import annotations

from typing import Mapping, Sequence

from lenskit.data import ID, GenericKey, ItemList, ItemListCollection, UserIDKey
from lenskit.pipeline import Pipeline

from ._results import BatchResults
from ._runner import BatchPipelineRunner, InvocationSpec

__all__ = ["BatchPipelineRunner", "BatchResults", "InvocationSpec", "predict", "recommend"]


def predict(
    pipeline: Pipeline,
    test: ItemListCollection[GenericKey] | Mapping[ID, ItemList],
    *,
    n_jobs: int | None = None,
    **kwargs,
) -> ItemListCollection[GenericKey]:
    """
    Convenience function to batch-generate rating predictions (or other per-item
    scores) from a pipeline.  This is a batch version of :func:`lenskit.predict`.

    Stability:
        Caller
    """

    runner = BatchPipelineRunner(n_jobs=n_jobs)
    runner.predict()
    outs = runner.run(pipeline, test)
    return outs.output("predictions")  # type: ignore


def score(
    pipeline: Pipeline,
    test: ItemListCollection[GenericKey] | Mapping[ID, ItemList],
    *,
    n_jobs: int | None = None,
    **kwargs,
) -> ItemListCollection[GenericKey]:
    """
    Convenience function to batch-generate personalized scores from a pipeline.
    This is a batch version of :func:`lenskit.predict`.

    Stability:
        Caller
    """

    runner = BatchPipelineRunner(n_jobs=n_jobs)
    runner.score()
    outs = runner.run(pipeline, test)
    return outs.output("scores")  # type: ignore


def recommend(
    pipeline: Pipeline,
    users: Sequence[ID | UserIDKey],
    n: int | None = None,
    candidates=None,
    *,
    n_jobs: int | None = None,
    **kwargs,
) -> ItemListCollection[UserIDKey]:
    """
    Convenience function to batch-generate recommendations from a pipeline. This
    is a batch version of :func:`lenskit.recommend`.

    .. todo::

        Support more inputs than just user IDs.

    Stability:
        Caller
    """

    runner = BatchPipelineRunner(n_jobs=n_jobs)
    runner.recommend(n=n)
    outs = runner.run(pipeline, users)
    return outs.output("recommendations")  # type: ignore



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Pipeline runner logic.
"""

# pyright: strict
from dataclasses import dataclass
from typing import Any, Generic, Literal, TypeAlias, TypeVar, get_args, get_origin

import structlog

from lenskit.diagnostics import PipelineError
from lenskit.logging import get_logger, trace

from ._impl import Pipeline
from .components import PipelineFunction
from .nodes import ComponentNode, InputNode, LiteralNode, Node
from .types import Lazy, is_compatible_data

_log = get_logger(__name__)
T = TypeVar("T")
State: TypeAlias = Literal["pending", "in-progress", "finished", "failed"]


class PipelineRunner:
    """
    Node status and results for a single pipeline run.

    This class operates recursively; pipelines should never be so deep that
    recursion fails.

    Stability:
        Internal
    """

    log: structlog.stdlib.BoundLogger
    pipe: Pipeline
    inputs: dict[str, Any]
    status: dict[str, State]
    state: dict[str, Any]

    def __init__(self, pipe: Pipeline, inputs: dict[str, Any]):
        self.log = _log.bind(pipeline=pipe.name)
        self.pipe = pipe
        self.inputs = inputs
        self.status = {n.name: "pending" for n in pipe.nodes}
        self.state = {}

    def run(self, node: Node[Any], *, required: bool = True) -> Any:
        """
        Run the pipleline to obtain the results of a node.
        """
        log = _log.bind(node=node.name)
        status = self.status[node.name]
        if status == "finished":
            return self.state[node.name]
        elif status == "in-progress":
            raise PipelineError(f"pipeline cycle encountered at {node}")
        elif status == "failed":  # pragma: nocover
            raise RuntimeError(f"{node} previously failed")

        trace(log, "processing node")
        self.status[node.name] = "in-progress"
        try:
            self._run_node(node, required)
            self.status[node.name] = "finished"
        except Exception as e:
            log.error("failed to run node", exc_info=e)
            self.status[node.name] = "failed"
            raise e

        try:
            return self.state[node.name]
        except KeyError as e:
            if required:
                raise e
            else:
                return None

    def _run_node(self, node: Node[Any], required: bool) -> None:
        match node:
            case LiteralNode(name, value):
                self.state[name] = value
            case InputNode(name, types=types):
                self._inject_input(name, types, required)
            case ComponentNode(name, comp, inputs, wiring):
                self._run_component(name, comp, inputs, wiring, required)
            case _:  # pragma: nocover
                raise PipelineError(f"invalid node {node}")

    def _inject_input(self, name: str, types: set[type] | None, required: bool) -> None:
        val = self.inputs.get(name, None)
        if val is None and required and types and not is_compatible_data(None, *types):
            raise PipelineError(f"input {name} not specified")

        if val is not None and types and not is_compatible_data(val, *types):
            raise TypeError(f"invalid data for input {name} (expected {types}, got {type(val)})")

        trace(self.log, "injecting input", name=name, value=val)
        self.state[name] = val

    def _run_component(
        self,
        name: str,
        comp: PipelineFunction[Any],
        inputs: dict[str, type | None],
        wiring: dict[str, str],
        required: bool,
    ) -> None:
        in_data = {}
        log = self.log.bind(node=name)
        trace(log, "processing inputs")
        for iname, itype in inputs.items():
            # look up the input wiring for this parameter input
            src = wiring.get(iname, None)
            if src is not None:
                snode = self.pipe.node(src)
            else:
                snode = self.pipe.get_default(iname)

            # check if this is a lazy node
            lazy = False
            if itype is not None and get_origin(itype) == Lazy:
                lazy = True
                (itype,) = get_args(itype)

            if snode is None:
                ival = None
            else:
                if required and itype:
                    ireq = not is_compatible_data(None, itype)
                else:
                    ireq = False

                if lazy:
                    ival = DeferredRun(self, iname, name, snode, required=ireq, data_type=itype)
                else:
                    ival = self.run(snode, required=ireq)

            # bail out if we're failing to satisfy a dependency but it is not required
            if (
                ival is None
                and itype
                and not lazy
                and not is_compatible_data(None, itype)
                and not required
            ):
                return None

            # check the data type before passing
            if itype and not lazy and not is_compatible_data(ival, itype):
                if ival is None:
                    raise TypeError(
                        f"no data available for required input ❬{iname}❭ on component ❬{name}❭"
                    )
                raise TypeError(
                    f"input ❬{iname}❭ on component ❬{name}❭ has invalid type {type(ival)} (expected {itype})"  # noqa: E501
                )

            in_data[iname] = ival

        trace(log, "running component", component=comp)
        self.state[name] = comp(**in_data)


@dataclass(eq=False)
class DeferredRun(Generic[T]):
    """
    Implementation of :class:`Lazy` for deferred runs in a pipeline runner.

    Stability:
        Internal
    """

    runner: PipelineRunner
    iname: str
    cname: str
    node: Node[T]
    required: bool
    data_type: type | None

    def get(self) -> T:
        val = self.runner.run(self.node, required=self.required)

        if self.data_type is not None and not is_compatible_data(val, self.data_type):
            raise TypeError(
                f"input ❬{self.iname}❭ on component ❬{self.cname}❭ has invalid type {type(val)} (expected {self.data_type})"  # noqa: E501
            )

        return val



"""
Extended logger providing TRACE support.
"""

from __future__ import annotations

import logging
import os
from typing import Any

import structlog

_trace_debug = os.environ.get("LK_TRACE", "no").lower() == "debug"


def trace(logger: structlog.stdlib.BoundLogger, *args: Any, **kwargs: Any):
    """
    Emit a trace-level message, if LensKit tracing is enabled.  Trace-level
    messages are more fine-grained than debug-level messages, and you usually
    don't want them.

    This function does not work on the lazy proxies returned by
    :func:`get_logger` and similar — it only works on bound loggers.

    Stability:
        Caller
    """
    meth = getattr(logger, "trace", None)
    if meth is not None:
        meth(*args, **kwargs)
    elif _trace_debug:
        logger.debug(*args, **kwargs)


class TracingLogger(structlog.stdlib.BoundLogger):
    """
    Class for LensKit loggers with trace-level logging support.
    """

    def bind(self, **new_values: Any) -> TracingLogger:
        return super().bind(**new_values)  # type: ignore[return-value]

    def new(self, **new_values: Any) -> TracingLogger:
        return super().new(**new_values)  # type: ignore[return-value]

    def trace(self, event: str | None, *args: Any, **kw: Any):
        if args:
            kw["positional_args"] = args
        try:
            args, kwargs = self._process_event("trace", event, kw)  # type: ignore
        except structlog.DropEvent:
            return None
        self._logger.debug(*args, **kwargs)


def lenskit_filtering_logger(level: int):
    """
    No-op filtering logger.
    """

    if level < logging.DEBUG:
        return TracingLogger

    def trace(self, event: str | None, *args: Any, **kw: Any):
        pass

    base = structlog.make_filtering_bound_logger(level)
    return type(f"LensKitLoggerFilter{level}", (base,), {"trace": trace})



"""
LensKit statistical computations.
"""

from __future__ import annotations

import warnings

import numpy as np
from numpy.typing import ArrayLike

from lenskit.diagnostics import DataWarning


def gini(xs: ArrayLike) -> float:
    """
    Compute the Gini coefficient of an array of values.

    This is inspired by `Olivia Guest's implementation`_ and based on the
    `StatsDirect reference`_.  It does *not* include Olivia's zero adjustment;
    the Gini coefficient is fine with some zeros, so long as the sum is strictly
    positive.

    .. _Olivia Guest's implementation: https://github.com/oliviaguest/gini
    .. _StatsDirect reference: https://www.statsdirect.com/help/default.htm#nonparametric_methods/gini.htm

    Stability:
        Caller

    Args:
        xs:
            An array of values representing the “resource” allocated to each
            instance.

    Returns:
        The Gini coefficient of the values in ``xs``.
    """
    xs = np.asarray(xs)
    if np.any(xs < 0):
        warnings.warn(
            "Gini coefficient is not defined for negative values", DataWarning, stacklevel=2
        )

    n = len(xs)
    xs = np.sort(xs)
    ranks = np.arange(1, n + 1, dtype=np.float64)
    ranks *= 2
    ranks -= n + 1
    num = np.sum(xs * ranks)
    denom = n * np.sum(xs, dtype=np.float64)
    if denom <= 0:
        warnings.warn(
            "Gini coefficient is not defined for non-positive totals", DataWarning, stacklevel=2
        )
    return max(num / denom, 0)


def argtopn(xs: ArrayLike, n: int) -> np.ndarray[int, np.dtype[np.int64]]:
    """
    Compute the ordered positions of the top *n* elements.  Similar to
    :func:`torch.topk`, but works with NumPy arrays and only returns the
    indices.
    """
    if n == 0:
        return np.empty(0, np.int64)

    xs = np.asarray(xs)

    N = len(xs)
    invalid = np.isnan(xs)
    if np.any(invalid):
        mask = ~invalid
        vxs = xs[mask]
        remap = np.arange(N)[mask]
        res = argtopn(vxs, n)
        return remap[res]

    if n >= 0 and n < N:
        parts = np.argpartition(-xs, n)
        top_scores = xs[parts[:n]]
        top_sort = np.argsort(-top_scores)
        order = parts[top_sort]
    else:
        order = np.argsort(-xs)

    return order



from __future__ import annotations

import logging

import numpy as np
from typing_extensions import override

from lenskit.data import Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_logger = logging.getLogger(__name__)


class TrainingCandidateSelectorBase(Component[ItemList], Trainable):
    """
    Base class for candidate selectors using the training data.

    Stability:
        Caller
    """

    config: None
    items_: Vocabulary
    """
    List of known items from the training data.
    """

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "items_") and not options.retrain:
            return

        self.items_ = data.items.copy()


class AllTrainingItemsCandidateSelector(TrainingCandidateSelectorBase):
    """
    Candidate selector that selects all known items from the training data.

    Stability:
        Caller
    """

    config: None

    def __call__(self) -> ItemList:
        return ItemList.from_vocabulary(self.items_)


class UnratedTrainingItemsCandidateSelector(TrainingCandidateSelectorBase):
    """
    Candidate selector that selects all known items from the training data that
    do not appear in the request user's history (:attr:`RecQuery.user_items`).
    If no item history is available, then all training items are returned.

    In order to look up the user's history in the training data, this needs to
    be combined with a component like
    :class:`~.history.UserTrainingHistoryLookup`.

    Stability:
        Caller
    """

    config: None

    def __call__(self, query: QueryInput) -> ItemList:
        query = RecQuery.create(query)
        items = ItemList.from_vocabulary(self.items_)

        if query.user_items is not None:
            mask = np.full(len(self.items_), True, np.bool_)
            qis = query.user_items.numbers(vocabulary=self.items_)
            qis = qis[qis >= 0]
            mask[qis] = False
            items = items[mask]

        return items



from __future__ import annotations

import logging
from datetime import datetime
from typing import Literal

import numpy as np
import pandas as pd
from pydantic import BaseModel
from typing_extensions import override

from lenskit.data import Dataset, ItemList, Vocabulary
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_log = logging.getLogger(__name__)


class PopConfig(BaseModel):
    """
    Configuration for popularity scoring.
    """

    score: Literal["quantile", "rank", "count"] = "quantile"
    """
    The method for computing popularity scores.  For all methods, higher scores
    represent more popular items.
    """


class PopScorer(Component[ItemList], Trainable):
    """
    Score items by their popularity.  Use with :py:class:`TopN` to get a
    most-popular-items recommender.

    Stability:
        Caller

    Attributes:
        item_pop_:
            Item popularity scores.
    """

    config: PopConfig

    items_: Vocabulary
    item_scores_: np.ndarray[int, np.dtype[np.float32]]

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "item_scores_") and not options.retrain:
            return

        _log.info("counting item popularity")
        self.items_ = data.items.copy()
        stats = data.item_stats()
        scores = stats["count"].reindex(self.items_.ids())
        self.item_scores_ = np.require(
            self._train_internal(scores).reindex(self.items_.ids()).values, np.float32
        )

    def _train_internal(self, scores: pd.Series) -> pd.Series:
        if self.config.score == "rank":
            _log.info("ranking %d items", len(scores))
            scores = scores.rank().sort_index()
        elif self.config.score == "quantile":
            _log.info("computing quantiles for %d items", len(scores))
            cmass = scores.sort_values()
            cmass = cmass.cumsum()
            cdens = cmass / scores.sum()
            scores = cdens.sort_index()
        elif self.config.score == "count":
            _log.info("scoring items with their rating counts")
            scores = scores.sort_index()
        else:
            raise ValueError("invalid scoring method " + repr(self.config.score))

        return scores

    def __call__(self, items: ItemList) -> ItemList:
        inums = items.numbers(vocabulary=self.items_, missing="negative")
        mask = inums >= 0
        scores = np.full(len(items), np.nan, np.float32)
        scores[mask] = self.item_scores_[inums[mask]]
        return ItemList(items, scores=scores)


class TimeBoundedPopConfig(PopConfig):
    cutoff: datetime
    """
    Time window for computing popularity scores.
    """


class TimeBoundedPopScore(PopScorer):
    """
    Score items by their time-bounded popularity, i.e., the popularity in the
    most recent `time_window` period.  Use with :py:class:`TopN` to get a
    most-popular-recent-items recommender.

    Attributes:
        item_scores_(pandas.Series):
            Time-bounded item popularity scores.
    """

    config: TimeBoundedPopConfig

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "item_scores_") and not options.retrain:
            return

        _log.info("counting time-bounded item popularity")
        log = data.interaction_log("numpy")

        item_scores = None
        if log.timestamps is None:
            _log.warning("no timestamps in interaction log; falling back to PopScorer")
            super().train(data, options)
            return
        else:
            counts = np.zeros(data.item_count, dtype=np.int32)
            start_timestamp = self.config.cutoff.timestamp()
            item_nums = log.item_nums[log.timestamps > start_timestamp]
            np.add.at(counts, item_nums, 1)

            item_scores = super()._train_internal(
                pd.Series(counts, index=data.items.index),
            )

        self.items_ = data.items.copy()
        self.item_scores_ = np.require(item_scores.reindex(self.items_.ids()).values, np.float32)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
from typing import Literal

from lenskit.data import ID, ItemList, RecQuery

from ._impl import Pipeline
from .components import Component


class RecPipelineBuilder:
    """
    Builder to help assemble common pipeline patterns.

    This is a convenience class; you can always directly assemble a
    :class:`Pipeline` if this class's behavior is inadquate.

    Stability:
        Caller
    """

    _selector: Component
    _scorer: Component
    _ranker: Component
    is_predictor: bool = False
    _predict_transform: Component | None = None
    _fallback: Component | None = None

    def __init__(self):
        from lenskit.basic.candidates import UnratedTrainingItemsCandidateSelector
        from lenskit.basic.topn import TopNRanker

        self._selector = UnratedTrainingItemsCandidateSelector()
        self._ranker = TopNRanker()

    def scorer(self, score: Component):
        """
        Specify the scoring model.
        """
        self._scorer = score

    def ranker(self, rank: Component | None = None, *, n: int = -1):
        """
        Specify the ranker to use.  If ``None``, sets up a :class:`TopNRanker`
        with ``n=n``.
        """
        from lenskit.basic.topn import TopNRanker

        if rank is None:
            self._ranker = TopNRanker(n=n)
        else:
            self._ranker = rank

    def candidate_selector(self, sel: Component):
        """
        Specify the candidate selector component.  The component should accept
        a query as its input and return an item list.
        """
        self._selector = sel

    def predicts_ratings(
        self, *, transform: Component | None = None, fallback: Component | None = None
    ):
        """
        Specify that this pipeline will predict ratings, optionally providing a
        rating transformer and fallback scorer for the rating predictions.

        Args:
            transform:
                A component to transform scores prior to returning them.  If
                supplied, it will be applied to both primary scores and fallback
                scores.
            fallback:
                A fallback scorer to use when the primary scorer cannot score an
                item. The fallback should accept ``query`` and ``items`` inputs,
                and return an item list.
        """
        self.is_predictor = True
        self._predict_transform = transform
        self._fallback = fallback

    def build(self, name: str | None = None) -> Pipeline:
        """
        Build the specified pipeline.
        """
        from lenskit.basic.composite import FallbackScorer
        from lenskit.basic.history import UserTrainingHistoryLookup

        pipe = Pipeline(name=name)

        query = pipe.create_input("query", RecQuery, ID, ItemList)

        items = pipe.create_input("items", ItemList)
        n_n = pipe.create_input("n", int, None)

        lookup = pipe.add_component("history-lookup", UserTrainingHistoryLookup(), query=query)
        cand_sel = pipe.add_component("candidate-selector", self._selector, query=lookup)
        candidates = pipe.use_first_of("candidates", items, cand_sel)

        n_score = pipe.add_component("scorer", self._scorer, query=lookup, items=candidates)
        if self.is_predictor:
            if self._fallback is not None:
                fb = pipe.add_component(
                    "fallback-predictor", self._fallback, query=lookup, items=candidates
                )
                rater = pipe.add_component(
                    "rating-merger", FallbackScorer(), scores=n_score, backup=fb
                )
            else:
                rater = n_score

            if self._predict_transform:
                pipe.add_component(
                    "rating-predictor", self._predict_transform, query=query, items=rater
                )
            else:
                pipe.alias("rating-predictor", rater)

        rank = pipe.add_component("ranker", self._ranker, items=n_score, n=n_n)
        pipe.alias("recommender", rank)

        return pipe


def topn_pipeline(
    scorer: Component,
    *,
    predicts_ratings: bool | Literal["raw"] = False,
    n: int = -1,
    name: str | None = None,
) -> Pipeline:
    """
    Create a pipeline that produces top-N recommendations using a scoring model.

    Stability:
        Caller

    Args:
        scorer:
            The scorer to use in the pipeline (it will added with the component
            name ``scorer``, see :ref:`pipeline-names`).
        predicts_ratings:
            If ``True``, make set up to predict ratings (``rating-predictor``),
            using ``scorer`` with a fallback of :class:`BiasScorer`; if
            ``"raw"``, use ``scorer`` directly with no fallback.
        n:
            The recommendation list length to configure in the pipeline.
        name:
            The pipeline name.
    """
    from lenskit.basic.bias import BiasScorer

    builder = RecPipelineBuilder()
    builder.scorer(scorer)
    builder.ranker(n=n)
    if predicts_ratings == "raw":
        builder.predicts_ratings()
    elif predicts_ratings:
        builder.predicts_ratings(fallback=BiasScorer())

    return builder.build(name)


def predict_pipeline(
    scorer: Component,
    *,
    fallback: bool | Component = True,
    n: int = -1,
    name: str | None = None,
) -> Pipeline:
    """
    Create a pipeline that predicts ratings, but does **not** include any
    ranking capabilities.  Mostly userful for testing and historical purposes.
    The resulting pipeline **must** be called with an item list.

    Stability:
        Caller

    Args:
        scorer:
            The scorer to use in the pipeline (it will added with the component
            name ``scorer``, see :ref:`pipeline-names`).
        fallback:
            Whether to use a fallback predictor when the scorer cannot score.
            When configured, the `scorer` node is the scorer, and the
            `rating-predictor` node applies the fallback.
        n:
            The recommendation list length to configure in the pipeline.
        name:
            The pipeline name.
    """
    from lenskit.basic.bias import BiasScorer
    from lenskit.basic.composite import FallbackScorer
    from lenskit.basic.history import UserTrainingHistoryLookup

    pipe = Pipeline(name=name)

    query = pipe.create_input("query", RecQuery, ID, ItemList)
    items = pipe.create_input("items", ItemList)

    lookup = pipe.add_component("history-lookup", UserTrainingHistoryLookup(), query=query)

    score = pipe.add_component("scorer", scorer, query=lookup, items=items)

    if fallback is True:
        fallback = BiasScorer()

    if fallback is False:
        pipe.alias("rating-predictor", score)
    else:
        backup = pipe.add_component("fallback-predictor", fallback, query=lookup, items=items)
        pipe.add_component("rating-predictor", FallbackScorer(), primary=score, fallback=backup)

    return pipe
